/*
    dlditool - Dynamically Linked Disk Interface patch tool
    Copyright (C) 2006  Michael Chisholm (Chishm)

	Send all queries to chishm@hotmail.com

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License along
    with this program; if not, write to the Free Software Foundation, Inc.,
    51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
*/

#include <stdio.h>
#include <malloc.h>
#include <string.h>

#ifndef bool
typedef enum {false = 0, true = !0} bool;
#endif

#define GBA_MEMOFFSET 0x08000000
#define MBGBA_MEMOFFSET 0x02000000
#define NDS_HEADER_SIZE 0x200
#define DSGBA_LOADER_SIZE 0x200
#define NDS_ARM9_SOURCE 0x020
#define NDS_ARM9_DEST 0x028

#define FEATURE_MEDIUM_CANREAD	0x00000001
#define FEATURE_MEDIUM_CANWRITE	0x00000002
#define FEATURE_SLOT_GBA	0x00000010
#define FEATURE_SLOT_NDS	0x00000020

#define MAGIC_TOKEN 0xBF8DA5ED

#define FIX_ALL	0x01
#define FIX_GLUE	0x02
#define FIX_GOT	0x04
#define FIX_BSS	0x08

#define DLDI_VERSION 1

enum FileType { FT_NDS, FT_DSGBA, FT_GBA, FT_MBGBA, FT_BIN };

struct DldiHeader {
	union {
		struct {
			unsigned int token;		// 0xBF8DA5ED
			char shortString[8];	// " Chishm"
		} magicParts;
		char magicString[12];		// "\xED\xA5\x8D\xBF Chishm"
	} identifier;
	unsigned char version;
	unsigned char driverSize;
	unsigned char fixSections;
	unsigned char allocatedSpace;

	char friendlyName[48];

	unsigned int __text_start;	// Data start
	unsigned int __data_end;	// Data end
	unsigned int __glue_start;	// Interworking glue start	-- Needs address fixing
	unsigned int __glue_end;	// Interworking glue end
	unsigned int __got_start;	// GOT start					-- Needs address fixing
	unsigned int __got_end;		// GOT end
	unsigned int __bss_start;	// bss start					-- Needs setting to zero
	unsigned int __bss_end;		// bss end

	// IO_INTERFACE data
	char ioType[4];
	unsigned int features;
	unsigned int startup;			
	unsigned int isInserted;		
	unsigned int readSectors;		
	unsigned int writeSectors;
	unsigned int clearStatus;
	unsigned int shutdown;
};



const char dldiMagicString[] = "\xED\xA5\x8D\xBF Chishm";

char *appFileData = NULL;
size_t appFileSize = 0;
char *dldiFileData = NULL;
size_t dldiFileSize = 0;

// --- stdio

int _memcmp(const char *s1, const char *s2, int n)
{
	while(--n && (*s1 == *s2))
	{
		s1++;
		s2++;
	}

	return (s1[0] - s2[0]);
}

void *_memset(void *dest, int v, int l)
{	char *d = (char *)dest;
	while(l--)
		*d++ = (char)v;
	return dest;
}

void *_memcpy(void *dst, const void *src, int l)
{
	char *d, *s;
	d = (char *)dst;
	s = (char *)src;
	while(l--)
		*d++ = *s++;
	return dst;
}

// ----------

#if 0

void printUsage (char* programName) {
	printf ("Usage:\n");
	printf ("%s [-t<type>|-s<offset>] <dldi> <app>\n", programName);
	printf ("   -t<type>      specifies the type of application <nds|dsgba|gba|mbgba|mb>\n");
	printf ("                 If not specified, then auto-detects based on extension\n");
	printf ("   -s<offset>    Hexadecimal address of the start offset of the application in RAM\n");
	printf ("                 Use this on raw binaries\n");
	printf ("   <dldi>        the dldi patch file to apply\n");
	printf ("   <app>         the application binary to apply the patch to\n");
	return;
}

#endif

#if 0

bool stringEndsWith (const char *str, const char *end) {
	const char* strEnd;
	if (strlen (str) < strlen(end)) {
		return false;
	}
	strEnd = &str[strlen (str) - strlen(end)];
	return (stricmp (strEnd, end) == 0);			
}

#endif

int quickFind (const char* data, const char* search, size_t dataLen, size_t searchLen) {
	const int* dataChunk = (const int*) data;
	int searchChunk = ((const int*)search)[0];
	int i;
	size_t dataChunkEnd = dataLen / sizeof(int);

	for ( i = 0; i < (int)dataChunkEnd; i++) {
		if (dataChunk[i] == searchChunk) {
			if ((i*sizeof(int) + searchLen) > dataLen) {
				return -1;
			}
			if (_memcmp (&data[i*sizeof(int)], search, searchLen) == 0) {
				return i*sizeof(int);
			}
		}
	}

	return -1;
}

#if 0

enum FileType detectFileType (const char* appFileName) {
	enum FileType fileType;
	if (stringEndsWith(appFileName, ".nds")) {
		fileType = FT_NDS;
	} else if (stringEndsWith(appFileName, ".nds.gba") || stringEndsWith(appFileName, ".ds.gba")) {
		fileType = FT_DSGBA;
	} else if (stringEndsWith(appFileName, ".mb.gba") || stringEndsWith(appFileName, ".mb")) {
		fileType = FT_MBGBA;
	} else if (stringEndsWith(appFileName, ".gba")) {
		fileType = FT_GBA;
	} else {
		fileType = FT_BIN;
	}			
	return fileType;
}

enum FileType specifiedFileType (const char* arg) {
	enum FileType fileType;
	if (stringEndsWith(arg, "nds")) {
		fileType = FT_NDS;
	} else if (stringEndsWith(arg, "ndsgba") || stringEndsWith(arg, "dsgba")) {
		fileType = FT_DSGBA;
	} else if (stringEndsWith(arg, "mbgba") || stringEndsWith(arg, "mb")) {
		fileType = FT_MBGBA;
	} else if (stringEndsWith(arg, "gba")) {
		fileType = FT_GBA;
	} else {
		fileType = FT_BIN;
	}			
	return fileType;
}

#endif

#if 0

int main(int argc, char* argv[])
{
	
	char *dldiFileName;
	char *appFileName;

	int memOffset = 0;			// Position of the loaded binary after loading into memory
	int fileOffset = 0;		// Position of the loaded binary within the file
	int patchOffset;		// Position of patch destination in the file
	int relocationOffset;	// Value added to all offsets within the patch to fix it properly
	int ddmemOffset;		// Original offset used in the DLDI file
	int ddmemStart;			// Start of range that offsets can be in the DLDI file
	int ddmemEnd;			// End of range that offsets can be in the DLDI file
	int ddmemSize;			// Size of range that offsets can be in the DLDI file

	int i;

	enum FileType fileType;

	FILE* dldiFile;
	FILE* appFile;

	struct DldiHeader *pDH;
	struct DldiHeader *pAH;
	int *pAppDldi;

	printf ("Dynamically Linked Disk Interface patch tool by Michael Chisholm (Chishm)\n\n");

	if (argc == 3) {
		dldiFileName = (char*)	malloc (strlen (argv[1]) + 1);
		strcpy (dldiFileName, argv[1]);
		appFileName = (char*)	malloc (strlen (argv[2]) + 1);
		strcpy (appFileName, argv[2]);

		fileType = detectFileType (appFileName);

	} else if (argc == 4) {
		dldiFileName = (char*)	malloc (strlen (argv[1]) + 1);
		strcpy (dldiFileName, argv[2]);
		appFileName = (char*)	malloc (strlen (argv[2]) + 1);
		strcpy (appFileName, argv[3]);	

		if (strstr (argv[1], "-t") == argv[1]) {
			// Specified a file type
			fileType = specifiedFileType (argv[1]);
		} else if (strstr (argv[1], "-s") == argv[1]) {
			fileType = FT_BIN;
			if (sscanf (argv[1], "-s%x",&memOffset) != 1) {
				printUsage (argv[0]);
				return -1;
			}
		}
	} else {
		printUsage(argv[0]);
		return -1;
	}

	if (!(dldiFile = fopen (dldiFileName, "rb"))) {
		printf ("File not found: %s\n", dldiFileName);
		return -1;
	}

	if (!(appFile = fopen (appFileName, "rb+"))) {
		printf ("File not found: %s\n", appFileName);
		return -1;
	}

	// Load the app file and the DLDI patch file
	fseek (appFile, 0, SEEK_END);
	appFileSize = ftell(appFile);
	appFileData = (char*) malloc (appFileSize);
	fseek (appFile, 0, SEEK_SET);

	fseek (dldiFile, 0, SEEK_END);
	dldiFileSize = ftell(dldiFile);
	dldiFileData = (char*) malloc (dldiFileSize);
	fseek (dldiFile, 0, SEEK_SET);

	if (!appFileData || !dldiFileData) {
		fclose (appFile);
		fclose (dldiFile);
		if (appFileData) free (appFileData);
		if (dldiFileData) free (dldiFileData);
		printf ("Out of memory\n");
		return -1;
	}

	fread (appFileData, 1, appFileSize, appFile);
	fread (dldiFileData, 1, dldiFileSize, dldiFile);
	fclose (dldiFile);

	// Get appropriate offsets for the start of the binary in the file and where it is loaded to in memory
	if (memOffset == 0) {
		switch (fileType) {
			case FT_NDS:
				fileOffset = ((int*)appFileData)[NDS_ARM9_SOURCE/sizeof(int)];
				memOffset = ((int*)appFileData)[NDS_ARM9_DEST/sizeof(int)];
				break;
			case FT_DSGBA:
				fileOffset = ((int*)appFileData)[(NDS_ARM9_SOURCE+DSGBA_LOADER_SIZE)/sizeof(int)] + DSGBA_LOADER_SIZE;
				memOffset = ((int*)appFileData)[(NDS_ARM9_DEST+DSGBA_LOADER_SIZE)/sizeof(int)];
				break;
			case FT_GBA:
				memOffset = GBA_MEMOFFSET;
				break;
			case FT_MBGBA:
				memOffset = MBGBA_MEMOFFSET;
				break;
			case FT_BIN:
			default:
				break;
		}
	}

	// Find the DSDI reserved space in the file
	patchOffset = quickFind (appFileData, dldiMagicString, appFileSize, sizeof(dldiMagicString)/sizeof(char));

	if (patchOffset < 0) {
		printf ("%s does not have a DLDI section\n", appFileName);
		return -1;
	}

	pDH = (struct DldiHeader *)dldiFileData;
	pAH = (struct DldiHeader *)&appFileData[patchOffset];
	pAppDldi = (int*)&appFileData[patchOffset];	// Access the in words

	// Make sure the DLDI file is valid and usable
	if (strcmp (dldiMagicString, pDH->identifier.magicString) != 0) {
		printf ("Invalid DLDI file\n");
		return -1;
	}
	if (pDH->version != DLDI_VERSION) {
		printf ("Incorrect DLDI file version. Expected %d, found %d.\n", DLDI_VERSION, pDH->version);
		return -1;
	}
	if (pDH->driverSize > pAH->allocatedSpace) {
		printf ("Not enough space for patch. Available %d bytes, need %d bytes\n", ( 1 << pAH->allocatedSpace), ( 1 << pDH->driverSize) );
		return -1;
	}

	ddmemOffset = pDH->__text_start;
	relocationOffset = patchOffset - fileOffset + memOffset - ddmemOffset;

	printf ("Old driver:          %s\n", pAH->friendlyName);
	printf ("New driver:          %s\n", pDH->friendlyName);
	printf ("\n");
	printf ("Position in file:    0x%08X\n", patchOffset);
	printf ("Position in memory:  0x%08X\n", memOffset);
	printf ("Patch base address:  0x%08X\n", ddmemOffset);
	printf ("Relocation offset:   0x%08X\n", relocationOffset);
	printf ("\n");

	ddmemStart = pDH->__text_start;
	ddmemSize = (1 << pDH->driverSize);
	ddmemEnd = ddmemStart + ddmemSize;

	// Remember how much space is actually reserved
	pDH->allocatedSpace = pAH->allocatedSpace;
	// Copy the DLDI patch into the application
	_memcpy (pAH, pDH, dldiFileSize);

	// Fix the section pointers in the header
	pAH->__text_start += relocationOffset;
	pAH->__data_end += relocationOffset;
	pAH->__glue_start += relocationOffset;
	pAH->__glue_end += relocationOffset;
	pAH->__got_start += relocationOffset;
	pAH->__got_end += relocationOffset;
	pAH->__bss_start += relocationOffset;
	pAH->__bss_end += relocationOffset;
	// Fix the function pointers in the header
	pAH->startup += relocationOffset;
	pAH->isInserted += relocationOffset;
	pAH->readSectors += relocationOffset;
	pAH->writeSectors += relocationOffset;
	pAH->clearStatus += relocationOffset;
	pAH->shutdown += relocationOffset;

	if (pDH->fixSections & FIX_ALL) { 
		// Search through and fix pointers within the data section of the file
		for (i = (pDH->__text_start - ddmemStart) / sizeof(int); i < (int)((pDH->__data_end - ddmemStart) / sizeof(int)); i++) {
			if ((ddmemStart <= pAppDldi[i]) && (pAppDldi[i] < ddmemEnd)) {
				pAppDldi[i] += relocationOffset;
			}
		}
	}

	if (pDH->fixSections & FIX_GLUE) { 
		// Search through and fix pointers within the glue section of the file
		for (i = (pDH->__glue_start - ddmemStart) / sizeof(int); i < (int)((pDH->__glue_end - ddmemStart) / sizeof(int)); i++) {
			if ((ddmemStart <= pAppDldi[i]) && (pAppDldi[i] < ddmemEnd)) {
				pAppDldi[i] += relocationOffset;
			}
		}
	}

	if (pDH->fixSections & FIX_GOT) { 
		// Search through and fix pointers within the Global Offset Table section of the file
		for (i = (pDH->__got_start - ddmemStart) / sizeof(int); i < (int)((pDH->__got_end - ddmemStart) / sizeof(int)); i++) {
			if ((ddmemStart <= pAppDldi[i]) && (pAppDldi[i] < ddmemEnd)) {
				pAppDldi[i] += relocationOffset;
			}
		}
	}

	if (pDH->fixSections & FIX_BSS) { 
		// Initialise the BSS to 0
		_memset (&pAppDldi [pDH->__bss_start - ddmemStart], 0, pDH->__bss_end - pDH->__bss_start);
	}

	// Write the patch back to the file
	fseek (appFile, patchOffset, SEEK_SET);
	fwrite (pAppDldi, 1, ddmemSize, appFile);
	fclose (appFile);

	free (appFileData);
	free (dldiFileData);

	printf ("Patched successfully\n");

	return 0;
}

#endif

extern int transmemory(char *_appFileData,size_t _appFileSize,char *_dldiFileData,size_t _dldiFileSize);

int transmemory(char *_appFileData,size_t _appFileSize,char *_dldiFileData,size_t _dldiFileSize)
{
	int memOffset = 0;			// Position of the loaded binary after loading into memory
	int fileOffset = 0;		// Position of the loaded binary within the file
	int patchOffset;		// Position of patch destination in the file
	int relocationOffset;	// Value added to all offsets within the patch to fix it properly
	int ddmemOffset;		// Original offset used in the DLDI file
	int ddmemStart;			// Start of range that offsets can be in the DLDI file
	int ddmemEnd;			// End of range that offsets can be in the DLDI file
	int ddmemSize;			// Size of range that offsets can be in the DLDI file

	int i;

	enum FileType fileType=FT_NDS;

	struct DldiHeader *pDH;
	struct DldiHeader *pAH;
	int *pAppDldi;

  appFileData=_appFileData;
  appFileSize=_appFileSize;
  dldiFileData=_dldiFileData;
  dldiFileSize=_dldiFileSize;
  
//	printf ("Dynamically Linked Disk Interface patch tool by Michael Chisholm (Chishm)\n\n");

	if (!appFileData || !dldiFileData) {
//		printf ("Out of memory\n");
		return -1;
	}

	// Get appropriate offsets for the start of the binary in the file and where it is loaded to in memory
	if (memOffset == 0) {
		switch (fileType) {
			case FT_NDS:
				fileOffset = ((int*)appFileData)[NDS_ARM9_SOURCE/sizeof(int)];
				memOffset = ((int*)appFileData)[NDS_ARM9_DEST/sizeof(int)];
				break;
			case FT_DSGBA:
				fileOffset = ((int*)appFileData)[(NDS_ARM9_SOURCE+DSGBA_LOADER_SIZE)/sizeof(int)] + DSGBA_LOADER_SIZE;
				memOffset = ((int*)appFileData)[(NDS_ARM9_DEST+DSGBA_LOADER_SIZE)/sizeof(int)];
				break;
			case FT_GBA:
				memOffset = GBA_MEMOFFSET;
				break;
			case FT_MBGBA:
				memOffset = MBGBA_MEMOFFSET;
				break;
			case FT_BIN:
			default:
				break;
		}
	}

	// Find the DSDI reserved space in the file
	patchOffset = quickFind (appFileData, dldiMagicString, appFileSize, sizeof(dldiMagicString)/sizeof(char));

	if (patchOffset < 0) {
//		printf ("does not have a DLDI section\n");
		return -1;
	}

	pDH = (struct DldiHeader *)dldiFileData;
	pAH = (struct DldiHeader *)&appFileData[patchOffset];
	pAppDldi = (int*)&appFileData[patchOffset];	// Access the in words

	// Make sure the DLDI file is valid and usable
	if (strcmp (dldiMagicString, pDH->identifier.magicString) != 0) {
//		printf ("Invalid DLDI file\n");
		return -1;
	}
	if (pDH->version != DLDI_VERSION) {
//		printf ("Incorrect DLDI file version. Expected %d, found %d.\n", DLDI_VERSION, pDH->version);
		return -1;
	}
	if (pDH->driverSize > pAH->allocatedSpace) {
//		printf ("Not enough space for patch. Available %d bytes, need %d bytes\n", ( 1 << pAH->allocatedSpace), ( 1 << pDH->driverSize) );
		return -1;
	}

	ddmemOffset = pDH->__text_start;
	relocationOffset = patchOffset - fileOffset + memOffset - ddmemOffset;

//	printf ("Old driver:          %s\n", pAH->friendlyName);
//	printf ("New driver:          %s\n", pDH->friendlyName);
//	printf ("\n");
//	printf ("Position in file:    0x%08X\n", patchOffset);
//	printf ("Position in memory:  0x%08X\n", memOffset);
//	printf ("Patch base address:  0x%08X\n", ddmemOffset);
//	printf ("Relocation offset:   0x%08X\n", relocationOffset);
//	printf ("\n");

	ddmemStart = pDH->__text_start;
	ddmemSize = (1 << pDH->driverSize);
	ddmemEnd = ddmemStart + ddmemSize;

	// Remember how much space is actually reserved
	pDH->allocatedSpace = pAH->allocatedSpace;
	// Copy the DLDI patch into the application
	_memcpy (pAH, pDH, dldiFileSize);

	// Fix the section pointers in the header
	pAH->__text_start += relocationOffset;
	pAH->__data_end += relocationOffset;
	pAH->__glue_start += relocationOffset;
	pAH->__glue_end += relocationOffset;
	pAH->__got_start += relocationOffset;
	pAH->__got_end += relocationOffset;
	pAH->__bss_start += relocationOffset;
	pAH->__bss_end += relocationOffset;
	// Fix the function pointers in the header
	pAH->startup += relocationOffset;
	pAH->isInserted += relocationOffset;
	pAH->readSectors += relocationOffset;
	pAH->writeSectors += relocationOffset;
	pAH->clearStatus += relocationOffset;
	pAH->shutdown += relocationOffset;

	if (pDH->fixSections & FIX_ALL) { 
		// Search through and fix pointers within the data section of the file
		for (i = (pDH->__text_start - ddmemStart) / sizeof(int); i < (int)((pDH->__data_end - ddmemStart) / sizeof(int)); i++) {
			if ((ddmemStart <= pAppDldi[i]) && (pAppDldi[i] < ddmemEnd)) {
				pAppDldi[i] += relocationOffset;
			}
		}
	}

	if (pDH->fixSections & FIX_GLUE) { 
		// Search through and fix pointers within the glue section of the file
		for (i = (pDH->__glue_start - ddmemStart) / sizeof(int); i < (int)((pDH->__glue_end - ddmemStart) / sizeof(int)); i++) {
			if ((ddmemStart <= pAppDldi[i]) && (pAppDldi[i] < ddmemEnd)) {
				pAppDldi[i] += relocationOffset;
			}
		}
	}

	if (pDH->fixSections & FIX_GOT) { 
		// Search through and fix pointers within the Global Offset Table section of the file
		for (i = (pDH->__got_start - ddmemStart) / sizeof(int); i < (int)((pDH->__got_end - ddmemStart) / sizeof(int)); i++) {
			if ((ddmemStart <= pAppDldi[i]) && (pAppDldi[i] < ddmemEnd)) {
				pAppDldi[i] += relocationOffset;
			}
		}
	}

	if (pDH->fixSections & FIX_BSS) { 
		// Initialise the BSS to 0
		_memset (&pAppDldi [pDH->__bss_start - ddmemStart], 0, pDH->__bss_end - pDH->__bss_start);
	}

	// Write the patch back to the file
	_memcpy(&appFileData[patchOffset],pAppDldi,ddmemSize);

//	printf ("Patched successfully\n");

	return 0;
}

