/********************************************************************
 *                                                                  *
 * THIS FILE IS PART OF THE OggVorbis 'TREMOR' CODEC SOURCE CODE.   *
 *                                                                  *
 * USE, DISTRIBUTION AND REPRODUCTION OF THIS LIBRARY SOURCE IS     *
 * GOVERNED BY A BSD-STYLE SOURCE LICENSE INCLUDED WITH THIS SOURCE *
 * IN 'COPYING'. PLEASE READ THESE TERMS BEFORE DISTRIBUTING.       *
 *                                                                  *
 * THE OggVorbis 'TREMOR' SOURCE CODE IS (C) COPYRIGHT 1994-2002    *
 * BY THE Xiph.Org FOUNDATION http://www.xiph.org/                  *
 *                                                                  *
 ********************************************************************

 function: simple example decoder using vorbisidec

 ********************************************************************/

/* Takes a vorbis bitstream from stdin and writes raw stereo PCM to
   stdout using vorbisfile. Using vorbisfile is much simpler than
   dealing with libvorbis. */

#pragma Ospace

#include <NDS.h>
#include <stdio.h>
#include <stdlib.h>

#include "plugin.h"
#include "plugin_def.h"

#include "_console.h"
#include "memtool.h"

#include "../../ipc3.h"

#include "plug_ogg.h"

#include "libogg/config_types.h"
#include "libogg/ivorbiscodec.h"
#include "libogg/ivorbisfile.h"

#pragma Ospace

//--------------------------------

#include "filesys.h"

static u32 fofs=0;

#pragma Otime
static size_t callbacks_read_func  (void *ptr, size_t size, size_t nmemb, void *datasource)
{
  int len;
  
  len=FileSys_OGG_fread(ptr,size*nmemb);
  fofs+=len;
  
  return(len);
}
#pragma Ospace

static int    callbacks_seek_func  (void *datasource, ogg_int64_t offset, int whence)
{
  switch(whence){
    case 0: { // set
      fofs=offset;
    } break;
    case 1: { // cur
      fofs+=offset;
    } break;
    case 2: { // end
      _consolePrintf("FileSize=%dbyte\n",IPC3->IR_filesize);
      fofs=IPC3->IR_filesize-offset;
    } break;
  }
    
  FileSys_OGG_fseek(fofs);
  return(0);
}

static int    callbacks_close_func (void *datasource)
{
  return(0);
}

static long   callbacks_tell_func  (void *datasource)
{
  return(fofs);
}

static ov_callbacks callbacks = {callbacks_read_func,callbacks_seek_func,callbacks_close_func,callbacks_tell_func};

//--------------------------------

//#define PCMReadBufCount (24576/6)
#define PCMReadBufCount (1024*3)
static s16 *PCMReadBuf=NULL;

typedef struct {
  OggVorbis_File vf;
  u32 SampleRate,ChannelCount;
} TOGGInfo;

static TOGGInfo OGGInfo;

//--------------------------------

bool StartOGG(void)
{
  fofs=0;
  PCMReadBuf=(s16*)safemalloc(PCMReadBufCount*4); // 16bit2ch
  
  {
    int ret=ov_open_callbacks((void*)1,&OGGInfo.vf,NULL,0,callbacks);
    
    switch(ret){
      case OV_EREAD: _consolePrint("ret=OVEREAD A read from media returned an error.\n"); break;
      case OV_ENOTVORBIS: _consolePrint("ret=OV_ENOTVORBIS Bitstream is not Vorbis data.\n"); break;
      case OV_EVERSION: _consolePrint("ret=OV_EVERSION Vorbis version mismatch.\n"); break;
      case OV_EBADHEADER: _consolePrint("ret=OV_EBADHEADER Invalid Vorbis bitstream header.\n"); break;
      case OV_EFAULT: _consolePrint("ret=OV_EFAULT Internal logic fault.\n"); break;
    }
    
    if(ret<0){
      _consolePrint("Input does not appear to be an Ogg bitstream.\n");
      FreeOGG();
      return(false);
    }
    
    _consolePrint("libogg initialized.\n");
  }
  
  {
/*
    char **ptr=ov_comment(&OGGInfo.vf,-1)->user_comments;
    
    if(*ptr!=NULL){
      _consolePrint("UserComments:\n");
      while(*ptr){
        _consolePrintf("%s\n",ptr);
        ++ptr;
      }
    }
*/
    
    _consolePrintf("Encoded by: %s\n",ov_comment(&OGGInfo.vf,-1)->vendor);
    _consolePrint("\n");
  }
  
  {
    vorbis_info *vi=ov_info(&OGGInfo.vf,-1);
    
    OGGInfo.SampleRate=vi->rate;
    OGGInfo.ChannelCount=vi->channels;
    
    _consolePrintf("SampleRate %dHz\n",OGGInfo.SampleRate);
    _consolePrintf("ChannelCount %dch\n",OGGInfo.ChannelCount);
    _consolePrintf("\n");
    if(vi->bitrate_lower!=0) _consolePrintf("bitrate lower =%dkbps\n",(s32)(vi->bitrate_lower/1000));
    if(vi->bitrate_nominal!=0) _consolePrintf("bitrate normal=%dkbps\n",(s32)(vi->bitrate_nominal/1000));
    if(vi->bitrate_upper!=0) _consolePrintf("bitrate upper =%dkbps\n",(s32)(vi->bitrate_upper/1000));
    if(vi->bitrate_window!=0) _consolePrintf("bitrate window=%dkbps\n",(s32)(vi->bitrate_window/1000));
    
    if((128000<vi->bitrate_lower)||(128000<vi->bitrate_nominal)||(128000<vi->bitrate_upper)){
      _consolePrint("\n");
      _consolePrint("bitrate exceeds 128kbps.\nPossibility with noise.\n");
    }
  }
  
//  _consolePrintf("warrning:In the OGG interface, there is a memory leak bug.\n");
  
  return(true);
}

#pragma Otime
u32 UpdateOGG(s16 *lbuf,s16 *rbuf)
{
  u32 unitsize;
  
  if(OGGInfo.ChannelCount==1){
    unitsize=2;
    }else{
    unitsize=4;
  }
  
  u32 SamplePos=0;
  
  {
    u8 *WriteBuf;
    int current_section;
    long ret;
    
    WriteBuf=(u8*)PCMReadBuf;
    WriteBuf+=SamplePos*unitsize;
    
    while(SamplePos<PCMReadBufCount){
      //PrfStart();
      ret=ov_read(&OGGInfo.vf,WriteBuf,(PCMReadBufCount-SamplePos)*unitsize,&current_section);
      /*
      static u32 tr=0,tt=0;
      tr+=ret;
      tt+=PrfEnd(ret);
      _consolePrintf("%dsample,%dus\n",tr,tt);
      */
      
      if(ret==0){
        _consolePrintf("Done.\n");
        break;
      }
      if(ret<0){
        _consolePrintf("ov_read() decode error:%d. skip.\n",ret);
      }
      if(ret<unitsize){
        _consolePrintf("ov_read() fatal error:%d\n",ret);
        while(1);
      }
      if(0<ret){
        WriteBuf+=ret;
        SamplePos+=ret/unitsize;
      }
    }
  }
  
  if(SamplePos>PCMReadBufCount){
    _consolePrintf("Fatal error. SamplePos=%d\n",SamplePos);
    _consolePrintf("Fatal error. PCMReadBufCount=%d\n",PCMReadBufCount);
    while(1);
  }
  
  if((lbuf!=NULL)&&(rbuf!=NULL)){
    if(OGGInfo.ChannelCount==1){
      s16 *pcmbuf=(s16*)PCMReadBuf;
      u32 idx;
      for(idx=SamplePos;idx>0;idx--){
        *lbuf++=*pcmbuf++;
      }
      }else{
/*
      s16 *pcmbuf=(s16*)PCMReadBuf;
      u32 idx;
      for(idx=SamplePos;idx>0;idx--){
        *lbuf++=*pcmbuf++;
        *rbuf++=*pcmbuf++;
      }
*/
      u32 *pcmbuf=(u32*)PCMReadBuf;
      u32 idx;
      for(idx=0;idx<SamplePos;idx++){
        u32 s=pcmbuf[idx];
        lbuf[idx]=(s16)(s>>16);
        rbuf[idx]=(s16)s;
      }
    }
  }
  
  return(SamplePos);
}
#pragma Ospace

void FreeOGG(void)
{
  _consolePrint("FreeOGG();\n");
  
  ov_clear(&OGGInfo.vf);
  
  if(PCMReadBuf!=NULL){
    safefree(PCMReadBuf);
    PCMReadBuf=NULL;
  }
  
  OGGInfo.SampleRate=0;
  OGGInfo.ChannelCount=0;
}

void OGG_SetFileOffset(u64 ofs)
{
  ov_pcm_seek(&OGGInfo.vf,ofs);
}

u32 OGG_GetChannelCount(void)
{
  return(OGGInfo.ChannelCount);
}

u32 OGG_GetSampleRate(void)
{
  return(OGGInfo.SampleRate);
}

u32 OGG_GetSamplePerFrame(void)
{
  return(PCMReadBufCount);
}

